/* pam_subid module */

/*
 * $Id: pam_subid.c 82 2007-01-26 06:36:10Z psnowber $
 *
 * Written by Andrew Morgan <morgan@parc.power.net> 1996/3/11
 * Adapted by Phil Snowberger <psnowber@cse.nd.edu> 2006-05-12
 *
 */

/*
 * here, we make definitions for the externally accessible functions
 * in this file (these definitions are required for static modules
 * but strongly encouraged generally) they are used to instruct the
 * modules include file to define their prototypes.
 */

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD

#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <security/pam_modules.h>
#include "util_uid.h"

/* --- authentication management functions --- */

PAM_EXTERN int
pam_sm_authenticate(pam_handle_t *pamh, int flags ,
		    int argc , const char **argv )
{
	int status;
	char* username = NULL;

	status = pam_get_user(pamh, (const char**) &username, NULL);
	if ( (status == PAM_CONV_ERR) || (status == PAM_CONV_AGAIN) ) {
		return PAM_USER_UNKNOWN;
	}
	if (username == NULL) {
		return PAM_SERVICE_ERR;
	}

	if (is_descendant(getuid(), getpwnam(username)->pw_uid)) {
		return PAM_SUCCESS;
	}

     return PAM_AUTH_ERR;
}

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh , int flags ,
	       int argc , const char **argv )
{
     return PAM_SUCCESS;
}

/* --- account management functions --- */

PAM_EXTERN int
pam_sm_acct_mgmt(pam_handle_t *pamh , int flags ,
		 int argc , const char **argv )
{
     return PAM_AUTH_ERR;
}

/* --- password management --- */

PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t *pamh , int flags ,
		 int argc , const char **argv )
{
     return PAM_AUTHTOK_ERR;
}

/* --- session management --- */

PAM_EXTERN int
pam_sm_open_session(pam_handle_t *pamh , int flags ,
		    int argc , const char **argv )
{
    return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_close_session(pam_handle_t *pamh , int flags ,
		     int argc , const char **argv )
{
     return PAM_SUCCESS;
}

/* end of module definition */

/* static module data */
#ifdef PAM_STATIC
struct pam_module _pam_deny_modstruct = {
    "pam_subid",
    pam_sm_authenticate,
    pam_sm_setcred,
    pam_sm_acct_mgmt,
    pam_sm_open_session,
    pam_sm_close_session,
    pam_sm_chauthtok
};
#endif
