/* $Id: subuseradd.c 84 2007-01-26 06:38:21Z psnowber $
 * vim: set tw=80 ts=4 sw=4 : */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>

#if 0
#define _GNU_SOURCE
#include <getopt.h>
#endif

#include "subid.h"
#include "util_uid.h"
#include "log.h"

void usage(int argc, char* argv[]) {
	printf("usage: %s <user name>\n", argv[0]);
	exit(1);
}

int use_fully_qualified = 0;
int home_in_home = 0;

int main(int argc, char* argv[]) {
	struct passwd* calling_pw = NULL;
	struct passwd* create_pw = NULL;
	pid_t childpid;
	int childstatus;
	char* nametoadd = NULL;
	char* callinghomedir = NULL;
	char* homedirtomake = NULL;

	int c;
	int _optind = -1;

	while ((c = getopt(argc, argv, "qh")) != -1) {
		switch(c) {
		case 'q':
			use_fully_qualified = 1;
			break;
		case 'h':
			home_in_home = 1;
			break;
		default:
			usage(argc, argv); /* calls exit(1) */
		}
	}
	_optind = optind; /* easier to watch this with gdb */

	if (_optind >= argc) {
		/* the user didn't specify a username */
		usage(argc, argv); /* calls exit(1) */
	}
	
	/* look up the username of the calling process with getuid() */
	{
		calling_pw = getpwuid(getuid());
		/* printf("username: %s\n",calling_pw->pw_name); */
		/* printf("gecos:    %s\n",calling_pw->pw_gecos); */
	}

	if (use_fully_qualified) {
		if (NULL != strstr(argv[_optind], USERNAME_SEPARATOR)) {
			/* USERNAME_SEPARATOR must not occur in argv[1] */
			fprintf(stderr, "bad username:  may not include the separator string '%s'\n", USERNAME_SEPARATOR);
			exit(1);
		}

		nametoadd = (char*)malloc(sizeof(char)*(strlen(calling_pw->pw_name) + 1 + strlen(argv[_optind]) + 1));
		if (!nametoadd) {
			fprintf(stderr, "couldn't allocate memory!");
			exit(1);
		}
		memset(nametoadd, 0, strlen(calling_pw->pw_name) + 1 + strlen(argv[_optind]) + 1);
		strcat(nametoadd, calling_pw->pw_name);
		strcat(nametoadd, USERNAME_SEPARATOR);
		strcat(nametoadd, argv[_optind]);
	} else {
		nametoadd = argv[_optind];
	}

	/* check to make sure the requested user id doesn't already exist */
	{
		errno = 0; /* says the getpwnam man page */
		create_pw = getpwnam(nametoadd);
		if (!create_pw) {
			/* careful -- we want the entry to not exist */
			switch (errno) {
				case ENOMEM:
				case EIO:
				case EINTR:
				case EMFILE:
				case ENFILE:
				case ERANGE:
					perror("main(): there was an error looking up that passwd entry: ");
					exit(127);
				default:
					errno = 0; /* NOOP */
					/*debug(1,"Okay, that user name doesn't exist yet\n");*/
			}
		} else { /* getpwnam returned a valid struct passwd* */
			printf("Cannot create subuser:  that user name already exists\n");
			exit(1);
		}

	}

	if (home_in_home) {
		callinghomedir = getenv("HOME");
		if (!callinghomedir) {
			fprintf(stderr,"calling user doesn't have a HOME\n");
			exit(1);
		}
		homedirtomake = (char *) malloc(sizeof(char)*(strlen(callinghomedir)+1+4+1+strlen(argv[_optind])+1));
		if (!homedirtomake) {
			fprintf(stderr,"couldn't allocate enough memory");
			exit(1);
		}
		memset(homedirtomake, 0, strlen(callinghomedir) + 1 + 4 + 1 + strlen(argv[_optind]) + 1);
		strcat(homedirtomake,callinghomedir);
		strcat(homedirtomake,"/home");
		errno = 0;
		if (!  mkdir(homedirtomake, 0755)) {
			switch (errno) {
				case EEXIST:
					break;
				case ENAMETOOLONG:
					fprintf(stderr, "the resulting directory name, \"%s\", is too long\n", homedirtomake);
					exit(1);
				default:
					perror("an error occurred while creating the home directory:");
					exit(1);
			}
		}
		strcat(homedirtomake,"/");
		strncat(homedirtomake,argv[_optind], 1024);
	} else {
		homedirtomake = (char *) malloc(sizeof(char)*(1+4+1+strlen(argv[_optind])+1));
		if (!homedirtomake) {
			fprintf(stderr,"couldn't allocate enough memory");
			exit(1);
		}
		memset(homedirtomake, 0, sizeof(char)*(1+4+1+strlen(argv[_optind])+1));
		strcat(homedirtomake,"/home/");
		strncat(homedirtomake,argv[_optind], 1024);
	}


	/* after here, we start actually affecting system files, so we need to be
	 * able to back out after any given step; enter the ugly goto section */
	childpid = fork();
	if (-1 == childpid) {
		/* error */
		perror("could not fork child process");
		goto fork_failure;
	} else if (0 == childpid) {
		/* child */

		/* append an entry to /etc/passwd */
		/* create the home directory */
		execl("/usr/sbin/useradd","/usr/sbin/useradd",nametoadd,"-m","-d",homedirtomake,(char*)NULL);
		perror("execl");
		goto fork_failure;
	}

	/* parent */
	waitpid(childpid, &childstatus, 0);
	if (!(WIFEXITED(childstatus) && (0 == WEXITSTATUS(childstatus)))) {
		/* the child failed for some reason */
		fprintf(stderr, "couldn't invoke /usr/sbin/useradd\n");
		goto fork_failure;
	}

	if (add_relation( getuid() /* parent */, 
                      getpwnam(nametoadd)->pw_uid /* child */)) {
		printf("couldn't write new subuser map file!\n");
		goto fork_failure;
	}

	/* the child exited correctly */

	goto success;
	
fork_failure:
	return 1;

success:
	log_fmt("user %s created a new sub-user %s", 
	        getpwuid(getuid())->pw_name, nametoadd);
	return 0;
}
