/* $Id: subuserchown.c 68 2007-01-22 03:58:56Z psnowber $
 * vim: set tw=80 ts=4 sw=4 : */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>

#include "subid.h"
#include "util_uid.h"
#include "log.h"

void usage(int argc, char* argv[]) {
	printf("usage: %s    <user name> <file>\n", argv[0]);
	printf("       %s -q <user name> <file>\n", argv[0]);
	exit(1);
}

int use_fully_qualified = 0;

int main(int argc, char* argv[]) {
	struct passwd* child_pw = NULL;
	struct passwd* file_pw = NULL;
	int i;
	struct stat s;
	char* from_name = NULL;
	char* to_name = NULL;

    int c;
    int _optind = -1;

    while ((c = getopt(argc, argv, "hq")) != -1) {
        switch(c) {
        case 'q':
            use_fully_qualified = 1;
            break;
        default:
            usage(argc, argv); /* calls exit(1) */
        }
    }
    _optind = optind; /* easier to watch this with gdb */

    if (_optind >= argc) {
        /* the user didn't specify a username */
        usage(argc, argv); /* calls exit(1) */
    }                                                                                    
	
	from_name = strdup(getpwuid(getuid())->pw_name);

	if (use_fully_qualified) {
		to_name = (char*) malloc(sizeof(char)*(strlen(from_name) + 1 + strlen(argv[_optind]) + 1));
		if (!to_name) {
			fprintf(stderr, "couldn't allocate memory!\n");
			exit(1);
		}
		memset(to_name, 0, strlen(from_name) + 1 + strlen(argv[_optind]) + 1);
		strcat(to_name, from_name);
		strcat(to_name, USERNAME_SEPARATOR);
		strcat(to_name, argv[_optind]);
	} else {
		to_name = argv[_optind];
	}

	errno = 0; /* says the getpwnam man page */
	child_pw = getpwnam(to_name);
	if (!child_pw) {
		switch (errno) {
			case ENOMEM:
			case EIO:
			case EINTR:
			case EMFILE:
			case ENFILE:
			case ERANGE:
				perror("main(): there was an error looking up that passwd entry");
				exit(127);
			default:
				fprintf(stderr,"the target user '%s' does not exist\n", to_name);
				exit(1);
		}
	} else { /* getpwnam returned a valid struct passwd* */
	}

	/* to proceed, the username to chown to must be a child of the caller, or
	 * the caller itself */
	/*if (! (is_child(username) || (getuid() == getpwnam(username)->pw_uid))) {*/
	/* if (!is_descendant(lines, from_name, to_name)) {*/
	/* TODO:  this is ugly, en-uid-ify the whole thing */
	if (!is_descendant(getpwnam(from_name)->pw_uid, getpwnam(to_name)->pw_uid)) {
		fprintf(stderr,"permission denied (maybe that user isn't inferior to you?)\n");
		return 1;
	}

	/* now we know that the requested user to change to is a subuser of the caller */
	/* so check that each file belongs to the calling user or a subuser thereof */

	for (i=_optind+1;i<argc;i++) {

		errno = 0;
		if (-1 == stat(argv[i], &s)) {
			perror("main: checking file owner");
			exit(1);
		}
		file_pw = getpwuid(s.st_uid);
	
		/* to proceed, the file must belong to a child, or to the caller */
		/*if (! (is_child(username) || (getuid() == s.st_uid))) { */
		/* if (!is_descendant(lines, from_name, file_pw->pw_name)) { */
		/* TODO:  this is ugly, en-uid-ify the whole thing */
		if (!is_descendant(getpwnam(from_name)->pw_uid, getpwnam(file_pw->pw_name)->pw_uid)) {
			fprintf(stderr,"permission denied (the file %s doesn't belong to you "
			        "or any of your subusers)\n",argv[i]);
			continue;
		}

		errno = 0;
		if (-1 == chown(argv[i], getpwnam(to_name)->pw_uid, -1)) {
			perror(argv[i]);
			continue;
		}

		log_fmt("user %s changed owner of file %s, previously owned by %s, "
		        "now owned by %s", from_name, argv[i], file_pw->pw_name, to_name);

	}
	return 0;
}
