/* $Id: subusersu.c 67 2007-01-22 03:25:10Z psnowber $
 * vim: set tw=80 ts=4 sw=4 : */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>

#include "subid.h"
#include "util_uid.h"
#include "log.h"

void usage(int argc, char* argv[]) {
	printf("usage: %s    <user name>\n", argv[0]);
	printf("       %s -q <user name>\n", argv[0]);
	exit(1);
}

/* TODO: should fully qualified names be the default or not? */
int use_fully_qualified = 0;

int main(int argc, char* argv[]) {
	struct passwd* child_pw = NULL;
	pid_t childpid;
	int childstatus;
	char* calling_name = NULL;
	char* name_to_su = NULL;

	int c;
    int _optind = -1;

    while ((c = getopt(argc, argv, "hq")) != -1) {
        switch(c) {
        case 'q':
            use_fully_qualified = 1;
            break;
        default:
            usage(argc, argv); /* calls exit(1) */
        }
    }
    _optind = optind; /* easier to watch this with gdb */

    if (_optind >= argc) {
        /* the user didn't specify a username */
        usage(argc, argv); /* calls exit(1) */
    }
	
	calling_name = getpwuid(getuid())->pw_name;

	if (use_fully_qualified) {
		name_to_su = (char*)malloc(sizeof(char)*(strlen(calling_name) + 1 + strlen(argv[_optind]) + 1));
		if (!name_to_su) {
			fprintf(stderr, "couldn't allocate memory!");
			exit(1);
		}
		memset(name_to_su, 0, strlen(calling_name) + 1 + strlen(argv[_optind]) + 1);
		strcat(name_to_su, calling_name);
		strcat(name_to_su, USERNAME_SEPARATOR);
		strcat(name_to_su, argv[_optind]);
	} else {
		name_to_su = argv[_optind];
	}
	
	/* check to make sure the requested user id doesn't already exist */
	{
		errno = 0; /* says the getpwnam man page */
		child_pw = getpwnam(name_to_su);
		if (!child_pw) {
			/* careful -- we want the entry to not exist */
			switch (errno) {
				case ENOMEM:
				case EIO:
				case EINTR:
				case EMFILE:
				case ENFILE:
				case ERANGE:
					perror("main(): there was an error looking up that passwd entry: ");
					exit(127);
				default:
					fprintf(stderr,"the target user '%s' does not exist\n",name_to_su);
					exit(1);
			}
		} else { /* getpwnam returned a valid (struct passwd *) */
		}

	}

	/* to proceed, the username to chown to must be a child of the caller, or
	 * the caller itself */
	/*if (! (is_child(username) || (getuid() == getpwnam(username)->pw_uid))) {*/
	/*if (!is_descendant(calling_name, name_to_su)) {*/
	/* TODO: this is ugly, en-uid-ify the whole thing */
	if (!is_descendant(getpwnam(calling_name)->pw_uid, getpwnam(name_to_su)->pw_uid)) {
		fprintf(stderr,"permission denied (maybe that user isn't inferior to you?)\n");
		return 1;
	}

	/* fork a child to run the chown */
	childpid = fork();
	if (-1 == childpid) {
		/* error */
		perror("could not fork child process");
		goto fork_failure;
	} else if (0 == childpid) {
		/* child */
		execl("/bin/su","-",name_to_su,NULL);
		perror("execl failed");
		goto fork_failure;
	}

	/* parent */
	/* this is here instead of at success: so that it will be logged now, and
	 * not when the subshell returns */
	log_fmt("user %s su'd to user %s",
			calling_name, name_to_su);

	waitpid(childpid, &childstatus, 0);
	if (!(WIFEXITED(childstatus) && (0 == WEXITSTATUS(childstatus)))) {
		/* the child failed for some reason */
		fprintf(stderr, "su failed\n");
		goto fork_failure;
	}

	/* the child exited correctly */

	goto success;

fork_failure:
	return 1;
	
success:
	return 0;
}
