/* $Id: subusersudo.c 70 2007-01-25 22:03:50Z psnowber $
 * vim: set tw=80 ts=4 sw=4 : */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>

#include "subid.h"
#include "util.h"

void usage(int argc, char* argv[]) {
	printf("usage: %s    <user name> <command>\n", argv[0]);
	printf("       %s -q <user name> <command>\n", argv[0]);
	exit(1);
}

int use_fully_qualified = 0;

int main(int argc, char* argv[]) {
	struct passwd* child_pw = NULL;
	char** lines = NULL;
	pid_t childpid;
	int childstatus;
	char** argvs = NULL;
	char sudopath[] = "/usr/bin/sudo";
	char arg1[] = "-u";
	char arg3[] = "-H";
	char arg4[] = "--";
	int i;
	char* calling_name = NULL;
	char* name_to_sudo = NULL;

	int c;
    int _optind = -1;

    while ((c = getopt(argc, argv, "hq")) != -1) {
        switch(c) {
        case 'q':
            use_fully_qualified = 1;
            break;
        default:
            usage(argc, argv); /* calls exit(1) */
        }
    }
    _optind = optind; /* easier to watch this with gdb */

    if (_optind >= argc) {
        /* the user didn't specify a username */
        usage(argc, argv); /* calls exit(1) */
    }                                                                                    

	calling_name = getpwuid(getuid())->pw_name;

#ifdef USE_FULLY_QUALIFIED
    name_to_sudo = (char*)malloc(sizeof(char)*(strlen(calling_name) + 1 + strlen(argv[_optind]) + 1));
    if (!name_to_sudo) {
        fprintf(stderr, "couldn't allocate memory!");
        exit(1);
    }
    memset(name_to_sudo, 0, strlen(calling_name) + 1 + strlen(argv[_optind]) + 1);
    strcat(name_to_sudo, calling_name);
    strcat(name_to_sudo, USERNAME_SEPARATOR);
    strcat(name_to_sudo, argv[_optind]);
#else
    name_to_sudo = argv[_optind];
#endif

	/* sudo argv[1] -H -- argv[2]..argv[n] */
	/* yes  yes     no no yes .. yes NULL */
	argvs = (char**) malloc (sizeof(char*) * (argc + 3));
	if (!argvs) {
		fprintf(stderr,"couldn't allocate memory\n");
		exit(1);
	}
	argvs[0] = sudopath;
	argvs[1] = arg1;
	argvs[2] = name_to_sudo;
	argvs[3] = arg3;
	argvs[4] = arg4;
	/* FIXME:  change i<3 to i<argc to support wildcards etc */
	for (i=0;i<argc-2;i++) {
		argvs[i+5] = argv[i+2];
	}
	argvs[i+5] = NULL;
	
	/* check to make sure the requested user id doesn't already exist */
	{
		errno = 0; /* says the getpwnam man page */
		child_pw = getpwnam(argv[1]);
		if (!child_pw) {
			/* careful -- we want the entry to not exist */
			switch (errno) {
				case ENOMEM:
				case EIO:
				case EINTR:
				case EMFILE:
				case ENFILE:
				case ERANGE:
					perror("main(): there was an error looking up that passwd entry: ");
					exit(127);
				default:
					fprintf(stderr,"the target user does not exist\n");
#ifndef TESTING
					exit(1);
#endif
			}
		} else { /* getpwnam returned a valid struct passwd* */
		}

	}

	lines = read_data();
	if (!lines) {
		fprintf(stderr,"couldn't read in subuser map file\n");
		exit(1);
	}

	/* to proceed, the username to sudo to must be a child of the caller, or
	 * the caller itself */
	/*if (! (is_child(username) || (getuid() == getpwnam(username)->pw_uid))) {*/
	if (!is_descendant(lines, calling_name, name_to_sudo)) {
		fprintf(stderr,"permission denied (maybe that user isn't inferior to you?)\n");
		return 1;
	}

	/* fork a child to run the chown */
	childpid = fork();
	if (-1 == childpid) {
		/* error */
		perror("could not fork child process");
		goto fork_failure;
	} else if (0 == childpid) {
		/* child */
		execv("/usr/bin/sudo",argvs);
		perror("execv failed");
		goto fork_failure;
	}

	/* parent */
	/* this is here rather than at success: so that it will be logged now,
	 * rather than when the sudo'd command returns */
	log_fmt("user %s sudo'd as user %s the following command (one arg per line)",
			calling_name, name_to_sudo);
	for (i=2;i<argc;i++) {
		log_fmt("%s", argv[i]);
	}
	log_fmt("end sudo command");

	waitpid(childpid, &childstatus, 0);
	if (!(WIFEXITED(childstatus) && (0 == WEXITSTATUS(childstatus)))) {
		/* the child failed for some reason */
		fprintf(stderr, "su failed\n");
		goto fork_failure;
	}

	/* the child exited correctly */

	goto success;

fork_failure:
	return 1;
	
success:
	return 0;
}
