/* $Id: util.c 40 2006-02-17 02:11:57Z psnowber $
 * vim: set tw=80 ts=4 sw=4 cindent : */

#include <stdio.h>
#include <stdlib.h>
#define __USE_GNU /* to get strndup */
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#include <stdarg.h>

#include "subid.h"
#include "util_uid.h"

static struct child_parent* new_cp() {
    struct child_parent* cp;
    cp = (struct child_parent*)malloc(sizeof(struct child_parent));
    if (!cp) {
        perror("Could not allocate memory\n");
        exit(1);
    }
    cp->child  = -1;
    cp->parent = -1;
    cp->next   = NULL;
	cp->frotzed = 0;

    return cp;
}

static void free_cp_list(struct child_parent* the_list) {
    struct child_parent* cp = the_list;
    struct child_parent* cp_old = NULL;

    while (NULL != cp) {
        cp_old = cp;
        cp = cp->next;
        free(cp_old);
    }
}

#define BUFFER_SIZE 1024
//void subidfs_read_data() {
struct child_parent* subid_read_data() {
    struct child_parent *head;
    struct child_parent *walker;
    char buffer[BUFFER_SIZE] = {0};
    FILE* sufile = NULL;
    int i;
    int j;
    // way, way overkill, but fie on str{tok,sep}.
    char parent_name[BUFFER_SIZE] = {0};
    char child_name[BUFFER_SIZE] = {0};

    // for getpwnam_r
    struct passwd pd;
    struct passwd* pwdptr = &pd;
    struct passwd* temp_pwd_ptr;
    char pwdbuffer[BUFFER_SIZE];
    int  pwdlinelen = sizeof(pwdbuffer);
    uid_t parent_uid;
	int fd = -1;
	int ret = -1;

    head = new_cp();
    head->child  = -2;
    head->parent = -2;
    head->next   = NULL;
	head->frotzed = 1;

    walker = head;

	errno = 0;
	fd = open(subusers_file, O_RDONLY);
	if (-1 == fd) {
		perror("subid_read_data: open");
		return NULL;
	}
	ret = flock(fd, LOCK_SH);
	if (-1 == ret) {
		perror("subid_read_data: flock");
		return NULL;
	}
    sufile = fdopen(fd, "r");
    if (!sufile) {
        perror("subid_read_data: fdopen");
		return NULL;
    }

    /* slurp in each line, allocating memory for each */
    while (fgets(buffer,BUFFER_SIZE,sufile)) {
        /* kill the newline that fgets leaves */
        if (buffer[strlen(buffer) - 1] != '\n') {
            fprintf(stderr, "last character of buffer was not a newline -- maybe BUFFER_SIZE isn't long enough?\n");
			return NULL;
        }
        buffer[strlen(buffer) - 1] = '\0';

        i = 0;
         while (buffer[i] != ':') {
            parent_name[i] = buffer[i];
            ++i;
        }
        parent_name[i] = '\0';
        ++i;

        if (0 != (getpwnam_r(parent_name,pwdptr,pwdbuffer,pwdlinelen,&temp_pwd_ptr))) {
            perror("getpwnam_r() error");
			return NULL;
        }
        parent_uid = pd.pw_uid;

        while (buffer[i] != '\0') {
            j = 0;
            while ((buffer[i] != ',') && (buffer[i] != '\0')) {
                child_name[j] = buffer[i];
                ++i;
                ++j;
            }
            child_name[j] = '\0';
            if (buffer[i] == '\0') {
                // undo the next incr, so that the enclosing while
                // doesn't walk off the end of the buffer
                --i;
            }
            ++i;

            if (0 != (getpwnam_r(child_name,pwdptr,pwdbuffer,pwdlinelen,&temp_pwd_ptr))) {
                perror("getpwnam_r() error");
				return NULL;
            }

            walker->next = new_cp();
            walker->next->child  = pd.pw_uid;
            walker->next->parent = parent_uid;
            walker = walker->next;
        }
    }
    fclose(sufile);

    return head;
}

static int subid_write_data(struct child_parent* list) {
	struct child_parent* p_walker = NULL;
	struct child_parent* c_walker = NULL;
	FILE* sufile = NULL;
	int fd = -1;
	int ret = -1;
	uid_t this_parent = -1;
	int need_comma = 0;

	errno = 0;
	fd = open(subusers_file, O_RDWR);
	if (-1 == fd) {
		perror("subid_write_data: open");
		exit(1);
	}
	ret = flock(fd, LOCK_EX);
	if (-1 == ret) {
		perror("subid_write_data: flock");
		exit(1);
	}
	ret = ftruncate(fd, 0);
	if (-1 == ret) {
		perror("subid_write_data: ftruncate");
		exit(1);
	}
    sufile = fdopen(fd, "w");
    if (!sufile) {
        perror("subid_write_data: fdopen");
		exit(1);
    }

	p_walker = list;
	while (p_walker) {
		if (p_walker->parent == -2 || p_walker->child == -2) {
			/* dummy node */
			p_walker = p_walker->next;
			continue;
		}
		if (p_walker->frotzed) {
			p_walker = p_walker->next;
			continue;
		}
		
		need_comma = 0;
		fprintf(sufile,"%s:",getpwuid(p_walker->parent)->pw_name);
		this_parent = p_walker->parent;
		c_walker = p_walker;
		while (c_walker) {
			if (c_walker->frotzed) {
				c_walker = c_walker->next;
				continue;
			}
			if (c_walker->parent == this_parent) {
				if (!need_comma) {
					need_comma = 1;
				} else {
					fprintf(sufile,",");
				}
				fprintf(sufile,"%s",getpwuid(c_walker->child)->pw_name);
				fflush(sufile);
				c_walker->frotzed = 1;
			}
			c_walker = c_walker->next;
		}
		fprintf(sufile,"\n");
	}

	fclose(sufile);

	return(0);
}

int add_relation(uid_t p, uid_t c) {
	struct child_parent* walker;
	struct child_parent* list = subid_read_data();
	struct child_parent* new  = new_cp();
	
	if (!list) {
		exit(1);
	}
	
	new->child  = c;
	new->parent = p;

	/* append new to list */
	walker = list;
	while (walker->next) {
		walker->frotzed = 0;
		walker = walker->next;
	}
	walker->next = new;
	
	return subid_write_data(list);
}

int remove_relation(uid_t c) {
	struct child_parent* walker;
	struct child_parent* list = subid_read_data();

	if (!list) {
		exit(1);
	}

	walker = list;
	while (walker) {
		if (walker->parent == c) {
			fprintf(stderr,"That user has sub-users, delete them first.\n");
			exit(1);
		}
		walker = walker->next;
	}
	
	/* easy case, the offending relation is at the head of the list */
	if (list->child == c) {
		return subid_write_data(list->next);
	}
	
	walker = list;
	while (walker->next) {
		if (walker->next->child == c) {
			/* found it, snip it out */
			walker->next = walker->next->next;
			return subid_write_data(list);
		}
		
		walker = walker->next;
	}
	
	fprintf(stderr,"Couldn't find that relation in /etc/subusers\n");
	exit(1);
}

static int check_child(struct child_parent* head, uid_t p, uid_t c) {
	struct child_parent* walker = head;

	if (p == c)
		return 1;

	for (walker = head; (NULL != walker); walker = walker->next) {
		if (walker->child == c) {
			if (walker->parent == p) {
				return 1;
			}
			return check_child(head, p, walker->parent);
		}
	}

	return 0;
}

//int is_strict_descendant(struct child_parent* head, uid_t p, uid_t c) {
int is_strict_descendant(uid_t p, uid_t c) {
	struct child_parent* list = subid_read_data();
	if (!list) {
		return 0;
	}
	return check_child(list, p, c);
}

//int is_descendant(struct child_parent* head, uid_t p, uid_t c) {
int is_descendant(uid_t p, uid_t c) {
	/* users are their own children */
	if (p == c) 
		return 1;
	struct child_parent* list = subid_read_data();
	if (!list) {
		return 0;
	}

	return check_child(list, p, c);
}

int is_strict_descendant_cached(struct child_parent* list, uid_t p, uid_t c) {
	if (!list) {
		return 0;
	}
	return check_child(list, p, c);
}
int is_descendant_cached(struct child_parent* list, uid_t p, uid_t c) {
	if (p == c) 
		return 1;
	if (!list) {
		return 0;
	}

	return check_child(list, p, c);
}
